"""
Core MS3Record implementation for pymseed

"""
import json
from typing import Optional, Any, Union, Callable
from .clib import clibmseed, ffi, cdata_to_string
from .util import nstime2timestr, timestr2nstime, encoding_string
from .exceptions import MiniSEEDError
from .definitions import TimeFormat, SubSecond


class MS3Record:
    """A miniSEED record wrapper around a MS3Record structure"""

    def __init__(
        self,
        reclen: Optional[int] = None,
        encoding: Optional[int] = None,
        samplecnt: Optional[int] = None,
        recordptr: Any = None,
    ) -> None:
        """
        Initialize MS3Record wrapper.

        Args:
            reclen: Record length in bytes (None for default)
            encoding: Data encoding format (None for default)
            samplecnt: Sample count (None for default)
            recordptr: Existing MS3Record pointer (for wrapping parsed records)
        """
        if recordptr is not None:
            # Wrap an existing record structure
            self._msr = recordptr
            self._msr_allocated = False
        else:
            # Allocate a new record
            self._msr = clibmseed.msr3_init(ffi.NULL)
            self._msr_allocated = True

            # Set values if provided
            if reclen is not None:
                self._msr.reclen = reclen
            if encoding is not None:
                self._msr.encoding = encoding
            if samplecnt is not None:
                self._msr.samplecnt = samplecnt

    def __del__(self) -> None:
        if self._msr and self._msr_allocated:
            msr_ptr = ffi.new("MS3Record **")
            msr_ptr[0] = self._msr
            clibmseed.msr3_free(msr_ptr)

    def __repr__(self) -> str:
        datasamples_str = "[]"
        if self._msr.numsamples > 0:
            samples = self.datasamples
            if samples:
                datasamples_str = (
                    str(samples[:5]) + " ..." if len(samples) > 5 else str(samples)
                )

        return (
            f"MS3Record(sourceid: {self.sourceid}\n"
            f"        pubversion: {self._msr.pubversion}\n"
            f"            reclen: {self._msr.reclen}\n"
            f"     formatversion: {self._msr.formatversion}\n"
            f"         starttime: {self._msr.starttime} => {self.starttime_str()}\n"
            f"         samplecnt: {self._msr.samplecnt}\n"
            f"          samprate: {self._msr.samprate}\n"
            f"             flags: {self._msr.flags} => {self.flags_dict()}\n"
            f"               CRC: {self._msr.crc} => {hex(self._msr.crc)}\n"
            f"          encoding: {self._msr.encoding} => {self.encoding_str()}\n"
            f"       extralength: {self._msr.extralength}\n"
            f"        datalength: {self._msr.datalength}\n"
            f"             extra: {self.extra}\n"
            f"        numsamples: {self._msr.numsamples}\n"
            f"       datasamples: {datasamples_str}\n"
            f"          datasize: {self._msr.datasize}\n"
            f"        sampletype: {self.sampletype} => {self.sampletype_str()}\n"
            f"    record pointer: {self._msr})"
        )

    def __lt__(self, obj: "MS3Record") -> bool:
        return (self.sourceid, self.starttime) < (obj.sourceid, obj.starttime)

    def __gt__(self, obj: "MS3Record") -> bool:
        return (self.sourceid, self.starttime) > (obj.sourceid, obj.starttime)

    def __le__(self, obj: "MS3Record") -> bool:
        return (self.sourceid, self.starttime) <= (obj.sourceid, obj.starttime)

    def __ge__(self, obj: "MS3Record") -> bool:
        return (self.sourceid, self.starttime) >= (obj.sourceid, obj.starttime)

    def __str__(self) -> str:
        return (
            f"{self.sourceid}, "
            f"{self.pubversion}, "
            f"{self.reclen}, "
            f"{self.samplecnt} samples, "
            f"{self.samprate} Hz, "
            f"{self.starttime_str()}"
        )

    @property
    def record(self) -> bytes:
        """Return raw, parsed miniSEED record as bytes (copy)"""
        if self._msr.record == ffi.NULL:
            raise ValueError("No raw record available")

        return bytes(ffi.buffer(self._msr.record, self._msr.reclen)[:])

    @property
    def record_mv(self) -> memoryview:
        """Return raw, parsed miniSEED record as memoryview (no copy)"""
        if self._msr.record == ffi.NULL:
            raise ValueError("No raw record available")

        return ffi.buffer(self._msr.record, self._msr.reclen)

    @property
    def reclen(self) -> int:
        """Return record length in bytes"""
        return self._msr.reclen

    @reclen.setter
    def reclen(self, value: int) -> None:
        """Set maximum record length in bytes"""
        self._msr.reclen = value

    @property
    def swapflag(self) -> int:
        """Return swap flags as raw integer"""
        return self._msr.swapflag

    def swapflag_dict(self) -> dict[str, bool]:
        """Return swap flags as dictionary"""
        swapflag = {}
        swapflag["header_swapped"] = bool(self._msr.swapflag & clibmseed.MSSWAP_HEADER)
        swapflag["payload_swapped"] = bool(self._msr.swapflag & clibmseed.MSSWAP_PAYLOAD)
        return swapflag

    @property
    def sourceid(self) -> Optional[str]:
        """Return source identifier as string"""
        return cdata_to_string(self._msr.sid)

    @sourceid.setter
    def sourceid(self, value: str) -> None:
        """Set source identifier

        The source identifier is limited to 63 characters.
        Typically this is an FDSN Source Identifier:
        https://docs.fdsn.org/projects/source-identifiers
        """
        if len(value) >= clibmseed.LM_SIDLEN:
            raise ValueError(f"Source ID too long (max {clibmseed.LM_SIDLEN-1} characters)")

        self._msr.sid = ffi.new(f"char[{clibmseed.LM_SIDLEN}]", value.encode("utf-8"))

    @property
    def formatversion(self) -> int:
        """Return format version"""
        return self._msr.formatversion

    @formatversion.setter
    def formatversion(self, value: int) -> None:
        """Set format version"""
        if value not in [2, 3]:
            raise ValueError(f"Invalid miniSEED format version: {value}")
        self._msr.formatversion = value

    @property
    def flags(self) -> int:
        """Return record flags as raw 8-bit integer"""
        return self._msr.flags

    @flags.setter
    def flags(self, value: int) -> None:
        """Set record flags as an 8-bit unsigned integer"""
        self._msr.flags = value

    def flags_dict(self) -> dict[str, bool]:
        """Return record flags as a dictionary"""
        flags = {}
        if self._msr.flags & 0x01:
            flags["calibration_signals_present"] = True
        if self._msr.flags & 0x02:
            flags["time_tag_is_questionable"] = True
        if self._msr.flags & 0x04:
            flags["clock_locked"] = True
        return flags

    @property
    def starttime(self) -> int:
        """Return start time as nanoseconds since Unix/POSIX epoch"""
        return self._msr.starttime

    @starttime.setter
    def starttime(self, value: int) -> None:
        """Set start time as nanoseconds since Unix/POSIX epoch"""
        self._msr.starttime = value

    @property
    def starttime_seconds(self) -> float:
        """Return start time as seconds since Unix/POSIX epoch"""
        return self._msr.starttime / clibmseed.NSTMODULUS

    @starttime_seconds.setter
    def starttime_seconds(self, value: float) -> None:
        """Set start time as seconds since Unix/POSIX epoch

        The value is limited to microsecond resolution and will be rounded
        to to ensure a consistent conversion to the internal representation.
        This is done to avoid floating point precision issues.
        """
        # Scale to microseconds, round to nearest integer, then scale to nanoseconds
        self._msr.starttime = int(value * 1000000 + 0.5) * 1000

    def starttime_str(
        self,
        timeformat: TimeFormat = TimeFormat.ISOMONTHDAY_Z,
        subsecond: SubSecond = SubSecond.NANO_MICRO_NONE,
    ) -> Optional[str]:
        """Return start time as formatted string"""
        return nstime2timestr(self._msr.starttime, timeformat, subsecond)

    def set_starttime_str(self, value: str) -> None:
        """Set start time from formatted string"""
        self._msr.starttime = timestr2nstime(value)

    @property
    def samprate(self) -> float:
        """Return nominal sample rate"""
        return self._msr.samprate

    @samprate.setter
    def samprate(self, value: float) -> None:
        """Set nominal sample rate

        For data sampled regularly, this value should be positive and represent
        samples/second (Hz).

        For data NOT sampled regularly, this value should be negative and
        represent the sample interval in seconds (-1/interval).
        """
        self._msr.samprate = value

    @property
    def samprate_raw(self) -> float:
        """Return sample rate for regular data or sample interval for irregular

        This directly returns the value from the structure, negative indicates
        the data are NOT regularly sampled and the value is an interval period.
        """
        return self._msr.samprate

    @property
    def encoding(self) -> int:
        """Return data encoding format code"""
        return self._msr.encoding

    @encoding.setter
    def encoding(self, value: int) -> None:
        """Set data encoding format code

        See DataEncoding enum for valid values
        """
        self._msr.encoding = value

    @property
    def pubversion(self) -> int:
        """Return publication version"""
        return self._msr.pubversion

    @pubversion.setter
    def pubversion(self, value: int) -> None:
        """Set publication version"""
        self._msr.pubversion = value

    @property
    def samplecnt(self) -> int:
        """Return number of samples specified in fixed header"""
        return self._msr.samplecnt

    @property
    def crc(self) -> int:
        """Return CRC of entire record"""
        return self._msr.crc

    @property
    def extralength(self) -> int:
        """Return length of extra headers in bytes"""
        return self._msr.extralength

    @property
    def datalength(self) -> int:
        """Return length of data payload in bytes"""
        return self._msr.datalength

    @property
    def extra(self) -> str:
        """Return extra headers as JSON string"""
        if self._msr.extra == ffi.NULL:
            return ""
        return cdata_to_string(self._msr.extra)

    @extra.setter
    def extra(self, value: str) -> None:
        """Set extra headers as JSON string, will be minified to reduce size"""
        if value:
            # Minify the JSON string to ensure valid JSON and minimize size
            minified = json.dumps(json.loads(value), separators=(",", ":"))

            c_value = ffi.new("char[]", minified.encode("utf-8"))
            status = clibmseed.mseh_replace(self._msr, c_value)
            if status < 0:
                raise ValueError(f"Error setting extra headers: {status}")

    @property
    def datasamples(self) -> memoryview:
        """Return data samples as a memoryview (no copy)

        A view of the data samples in an internal buffer owned by this MS3Record
        instance is returned.  If the data are needed beyond the lifetime of this
        instance, a copy must be made.

        The returned view can be used directly with slicing and indexing
        from `0` to `MS3Record.numsamples - 1`.

        The view can efficiently be copied to a _python list_ using:

            data_samples = MS3Record.datasamples[:]
        """
        if self._msr.numsamples <= 0:
            return memoryview(b'')  # Empty memoryview

        sampletype = self.sampletype

        if sampletype == "i":
            ptr = ffi.cast("int32_t *", self._msr.datasamples)
            buffer = ffi.buffer(ptr, self._msr.numsamples * ffi.sizeof("int32_t"))
            return memoryview(buffer).cast('i')
        elif sampletype == "f":
            ptr = ffi.cast("float *", self._msr.datasamples)
            buffer = ffi.buffer(ptr, self._msr.numsamples * ffi.sizeof("float"))
            return memoryview(buffer).cast('f')
        elif sampletype == "d":
            ptr = ffi.cast("double *", self._msr.datasamples)
            buffer = ffi.buffer(ptr, self._msr.numsamples * ffi.sizeof("double"))
            return memoryview(buffer).cast('d')
        elif sampletype == "t":
            ptr = ffi.cast("char *", self._msr.datasamples)
            buffer = ffi.buffer(ptr, self._msr.numsamples)
            return memoryview(buffer).cast('B')
        else:
            raise ValueError(f"Unknown sample type: {sampletype}")

    @property
    def np_datasamples(self) -> Any:
        """Return data samples as a numpy array view (no copy)

        A view of the data samples in an internal buffer owned by this MS3Record
        instance is returned. If the data are needed beyond the lifetime of this
        instance, a copy must be made.
        """
        try:
            import numpy as np
        except ImportError:
            raise ImportError(
                "numpy is not installed. Install numpy or this package with [numpy] optional dependency"
            ) from None

        if self._msr.numsamples <= 0:
            return np.array([])  # Empty array

        sampletype = self.sampletype

        # Translate libmseed sample type to numpy type
        nptype = {
            "i": np.int32,
            "f": np.float32,
            "d": np.float64,
            "t": "S1",  # 1-byte strings for text data
        }

        if sampletype not in nptype:
            raise ValueError(f"Unknown sample type: {sampletype}")

        # Create numpy array view from CFFI buffer
        return np.frombuffer(self.datasamples, dtype=nptype[sampletype])

    @property
    def datasize(self) -> int:
        """Return size of decoded data payload in bytes"""
        return self._msr.datasize

    @property
    def numsamples(self) -> int:
        """Return number of decoded samples at MS3Record.datasamples"""
        return self._msr.numsamples

    @property
    def sampletype(self) -> Optional[str]:
        """Return sample type code if available, otherwise None"""
        if self._msr.sampletype:
            return self._msr.sampletype.decode("ascii")
        else:
            return None

    def sampletype_str(self) -> Optional[str]:
        """Return sample type as descriptive string"""
        sampletype = self.sampletype
        if sampletype == "i":
            return "int32"
        elif sampletype == "f":
            return "float32"
        elif sampletype == "d":
            return "float64"
        elif sampletype == "t":
            return "text"
        else:
            return None

    @property
    def endtime(self) -> int:
        """Return end time as nanoseconds since Unix/POSIX epoch"""
        return clibmseed.msr3_endtime(self._msr)

    @property
    def endtime_seconds(self) -> float:
        """Return end time as seconds since Unix/POSIX epoch"""
        return clibmseed.msr3_endtime(self._msr) / clibmseed.NSTMODULUS

    def endtime_str(
        self,
        timeformat: TimeFormat = TimeFormat.ISOMONTHDAY_Z,
        subsecond: SubSecond = SubSecond.NANO_MICRO_NONE,
    ) -> Optional[str]:
        """Return end time as formatted string"""
        return nstime2timestr(self.endtime, timeformat, subsecond)

    def encoding_str(self) -> Optional[str]:
        """Return encoding format as descriptive string"""
        return encoding_string(self._msr.encoding)

    def print(self, details: int = 0) -> None:
        """Print details of the record to stdout, with varying levels of `details`"""
        clibmseed.msr3_print(self._msr, details)

    def _record_handler_wrapper(self, record: Any, record_length: int, handlerdata: Any) -> None:
        """Callback function for msr3_pack()"""
        # Convert CFFI buffer to bytes for the handler
        record_bytes = ffi.buffer(record, record_length)[:]
        self._record_handler(record_bytes, self._record_handler_data)

    def pack(
        self,
        handler: Callable[[bytes, Any], None],
        handler_data: Any = None,
        data_samples: Optional[Union[list[int], list[float], list[str]]] = None,
        sample_type: Optional[str] = None,
        verbose: int = 0,
    ) -> tuple[int, int]:
        """Pack `datasamples` into miniSEED record(s) and call `handler()`

        The `handler(record, handlerdata)` function must accept two arguments:
                record:         A buffer containing a miniSEED record
                handlerdata:    The `handlerdata` value

        The handler function must use or copy the record buffer as the memory may be
        reused on subsequent iterations.

        Returns a tuple of (packed_samples, packed_records)
        """
        # Set handler function as CFFI callback function
        self._record_handler = handler
        self._record_handler_data = handler_data

        # Create callback function type and instance
        RECORD_HANDLER = ffi.callback(
            "void(char *, int, void *)", self._record_handler_wrapper
        )

        packed_samples = ffi.new("int64_t *")
        flags = clibmseed.MSF_FLUSHDATA  # Always flush data when packing

        if data_samples is not None and sample_type is not None:
            msr_datasamples = self._msr.datasamples
            msr_sampletype = self._msr.sampletype
            msr_numsamples = self._msr.numsamples
            msr_samplecnt = self._msr.samplecnt

            # Set up data samples in the MS3Record before packing
            self._msr.samplecnt = len(data_samples)
            self._msr.numsamples = len(data_samples)

            # Set sample type (use first character only)
            self._msr.sampletype = sample_type[0].encode('ascii')

            # Allocate and copy data samples based on type
            if sample_type == "i":
                try:
                    mv = memoryview(data_samples)
                    if mv.format == 'i' and mv.itemsize == 4:
                        # Compatible format - safe to zero-copy
                        sample_array = ffi.cast("int32_t *", ffi.from_buffer(data_samples))
                    else:
                        raise ValueError("Incompatible buffer format")
                except (TypeError, ValueError):
                    # Not compatible or not a buffer - need conversion
                    sample_array = ffi.new("int32_t[]", [int(sample) for sample in data_samples])
                self._msr.datasamples = ffi.cast("void *", sample_array)
                self._msr.datasize = len(data_samples) * 4
            elif sample_type == "f":
                try:
                    mv = memoryview(data_samples)
                    if mv.format == 'f' and mv.itemsize == 4:
                        # Compatible format - safe to zero-copy
                        sample_array = ffi.cast("float *", ffi.from_buffer(data_samples))
                    else:
                        raise ValueError("Incompatible buffer format")
                except (TypeError, ValueError):
                    # Not compatible or not a buffer - need conversion
                    sample_array = ffi.new("float[]", [float(sample) for sample in data_samples])
                self._msr.datasamples = ffi.cast("void *", sample_array)
                self._msr.datasize = len(data_samples) * 4
            elif sample_type == "d":
                try:
                    mv = memoryview(data_samples)
                    if mv.format == 'd' and mv.itemsize == 8:
                        # Compatible format - safe to zero-copy
                        sample_array = ffi.cast("double *", ffi.from_buffer(data_samples))
                    else:
                        raise ValueError("Incompatible buffer format")
                except (TypeError, ValueError):
                    # Not compatible or not a buffer - need conversion
                    sample_array = ffi.new("double[]", [float(sample) for sample in data_samples])
                self._msr.datasamples = ffi.cast("void *", sample_array)
                self._msr.datasize = len(data_samples) * 8
            elif sample_type == "t":
                try:
                    mv = memoryview(data_samples)
                    if mv.format in ('c', 'b', 'B') and mv.itemsize == 1:
                        # Compatible format - safe to zero-copy
                        sample_array = ffi.cast("char *", ffi.from_buffer(data_samples))
                    else:
                        raise ValueError("Incompatible buffer format")
                except (TypeError, ValueError):
                    # Not compatible or not a buffer - need conversion
                    text_data = []
                    for sample in data_samples:
                        if isinstance(sample, str):
                            text_data.append(sample.encode('utf-8')[0])
                        else:
                            text_data.append(int(sample) if isinstance(sample, (int, float)) else str(sample).encode('utf-8')[0])
                    sample_array = ffi.new("char[]", text_data)
                self._msr.datasamples = ffi.cast("void *", sample_array)
                self._msr.datasize = len(data_samples)
            else:
                raise ValueError(f"Unknown sample type: {sample_type}")

        packed_records = clibmseed.msr3_pack(
            self._msr,
            RECORD_HANDLER,
            ffi.NULL,
            packed_samples,
            flags,
            verbose,
        )

        # Restore original values if they were replaced
        if data_samples is not None and sample_type is not None:
            self._msr.datasamples = msr_datasamples
            self._msr.sampletype = msr_sampletype
            self._msr.numsamples = msr_numsamples
            self._msr.samplecnt = msr_samplecnt

        if packed_records < 0:
            raise MiniSEEDError(packed_records, "Error packing miniSEED record(s)")

        return (packed_samples[0], packed_records)



#!/usr/bin/env python3
#
# Read miniSEED file(s) using the pymseed package and decode
# data samples directly into NumPy arrays.
#
# The data are organized into a list of dictionaries, one for
# each contiguous trace of data.  The dictionaries contain very
# basic metadata and a NumPy array of data samples.  This data
# structure is for illustration only, and would like need
# adaptation for a real application.
#
# This file is part of the pymseed package.
# Copyright (c) 2025, EarthScope Data Services

import os
import sys
import pprint
from pymseed import MS3TraceList, sourceid2nslc

input_files = []

# Verify that input files are readable
for arg in sys.argv[1:]:
    if os.access(arg, os.R_OK):
        input_files.append(arg)
    else:
        sys.exit("Cannot read file: %s" % arg)

if not input_files:
    sys.exit("No input files specified")

# List of dictionaries for each trace
traces = []

mstl = MS3TraceList()

# Read all input files, creating record lists and _not_ unpacking data samples
for file in input_files:
    print("Reading file: %s" % file)
    mstl.read_file(file, unpack_data=False, record_list=True)

for traceid in mstl.traceids():
    for segment in traceid.segments():

        # Create a numpy array containing the data samples of this segment
        data_samples = segment.create_numpy_array_from_recordlist()

       # Create a dictionary for the trace with basic metadata
        trace = {
            "sourceid": traceid.sourceid,
            "NSLC": sourceid2nslc(traceid.sourceid),
            "publication_version": traceid.pubversion,
            "start_time": segment.starttime_str(),
            "end_time": segment.endtime_str(),
            "sample_rate": segment.samprate,
            "data_samples": data_samples,
        }

        traces.append(trace)

# Pretty print the trace list
pp = pprint.PrettyPrinter(indent=4, sort_dicts=False)
pp.pprint(traces)



"""
Core MS3RecordBufferReader implementation for pymseed

"""
from typing import Any, Optional
from .clib import clibmseed, ffi
from .msrecord import MS3Record
from .exceptions import MiniSEEDError


class MS3RecordBufferReader:
    """Read miniSEED records from a buffer, i.e. bytearray or numpy.array

    The `source` object must support the buffer protocol, e.g. bytearray,
    bytes, memoryview, numpy.array, etc.  This class will not modify the
    source buffer.

    If `unpack_data` is True, the data samples will be decoded.

    If `validate_crc` is True, the CRC will be validated if contained in
    the record (legacy miniSEED v2 contains no CRCs).  The CRC provides an
    internal integrity check of the record contents.
    """

    def __init__(
        self,
        source: Any,
        unpack_data: bool = False,
        validate_crc: bool = True,
        verbose: int = 0,
    ) -> None:
        self._msr_ptr = ffi.new("MS3Record **")
        self.source = ffi.from_buffer(source)
        self.source_offset = 0
        self.verbose = verbose

        # Construct parse flags
        self.parse_flags = 0
        if unpack_data:
            self.parse_flags |= clibmseed.MSF_UNPACKDATA
        if validate_crc:
            self.parse_flags |= clibmseed.MSF_VALIDATECRC

    def __enter__(self) -> "MS3RecordBufferReader":
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        self.close()

    def __iter__(self) -> "MS3RecordBufferReader":
        return self

    def __next__(self) -> MS3Record:
        next = self.read()
        if next is not None:
            return next
        else:
            raise StopIteration

    def read(self) -> Optional[MS3Record]:
        """Read the next miniSEED record from the buffer"""
        remaining_bytes = len(self.source) - self.source_offset
        if remaining_bytes < clibmseed.MINRECLEN:
            return None

        status = clibmseed.msr3_parse(
            self.source + self.source_offset,
            remaining_bytes,
            self._msr_ptr,
            self.parse_flags,
            self.verbose,
        )

        if status == clibmseed.MS_NOERROR:
            self.source_offset += self._msr_ptr[0].reclen
            return MS3Record(recordptr=self._msr_ptr[0])
        elif status > 0:  # Record detected but not enough data
            return None
        else:
            raise MiniSEEDError(status, "Error reading miniSEED record")

    def close(self) -> None:
        """Close the reader and free any allocated memory"""
        if self._msr_ptr[0] != ffi.NULL:
            clibmseed.msr3_free(self._msr_ptr)
            self._msr_ptr[0] = ffi.NULL



"""
Core file reader implementation for pymseed

"""
from typing import Optional, Union, Any

from .clib import clibmseed, ffi, cdata_to_string
from .exceptions import MiniSEEDError
from .msrecord import MS3Record


class MS3RecordReader:
    """Read miniSEED records from a file or file descriptor

    If `input` is an integer, it is assumed to be an open file descriptor,
    otherwise it is assumed to be a path (file) name.  In all cases the
    file or descriptor will be closed when the object's close() is called.

    If `unpack_data` is True, the data samples will be decoded.

    If `skip_not_data` is True, bytes from the input stream will be skipped
    until a record is found.

    If `validate_crc` is True, the CRC will be validated if contained in
    the record (legacy miniSEED v2 contains no CRCs).  The CRC provides an
    internal integrity check of the record contents.
    """

    def __init__(
        self,
        input: Union[str, int],
        unpack_data: bool = False,
        skip_not_data: bool = False,
        validate_crc: bool = True,
        verbose: int = 0,
    ) -> None:
        self._msfp_ptr = ffi.new("MS3FileParam **")
        self._msr_ptr = ffi.new("MS3Record **")
        self._selections = ffi.NULL
        self.verbose = verbose

        # Construct parse flags
        self.parse_flags = 0
        if unpack_data:
            self.parse_flags |= clibmseed.MSF_UNPACKDATA
        if skip_not_data:
            self.parse_flags |= clibmseed.MSF_SKIPNOTDATA
        if validate_crc:
            self.parse_flags |= clibmseed.MSF_VALIDATECRC

        # If the stream is an integer, assume an open file descriptor
        if isinstance(input, int):
            self._msfp_ptr[0] = clibmseed.ms3_mstl_init_fd(input)

            if self._msfp_ptr[0] == ffi.NULL:
                raise MiniSEEDError(
                    clibmseed.MS_GENERROR,
                    f"Error initializing file descriptor {input}",
                )

            self.stream_name = ffi.new("char[]", f"File Descriptor {input}".encode("utf-8"))
        # Otherwise, assume a path name
        else:
            self.stream_name = ffi.new("char[]", input.encode("utf-8"))

    def __enter__(self) -> "MS3RecordReader":
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        self.close()

    def __iter__(self) -> "MS3RecordReader":
        return self

    def __next__(self) -> MS3Record:
        next = self.read()
        if next is not None:
            return next
        else:
            raise StopIteration

    def read(self) -> Optional[MS3Record]:
        """Read the next miniSEED record from the file/descriptor"""

        status = clibmseed.ms3_readmsr_selection(
            self._msfp_ptr,
            self._msr_ptr,
            self.stream_name,
            self.parse_flags,
            self._selections,
            self.verbose,
        )

        if status == clibmseed.MS_NOERROR:
            return MS3Record(recordptr=self._msr_ptr[0])
        elif status == clibmseed.MS_ENDOFFILE:
            return None
        else:
            raise MiniSEEDError(status, "Error reading miniSEED record")

    def close(self) -> None:
        """Close the reader and free any allocated memory"""

        # Perform cleanup by calling the function with NULL stream name
        if self._msfp_ptr[0] != ffi.NULL or self._msr_ptr[0] != ffi.NULL:
            clibmseed.ms3_readmsr_selection(
                self._msfp_ptr,
                self._msr_ptr,
                ffi.NULL,  # NULL stream name signals cleanup
                self.parse_flags,
                self._selections,
                self.verbose,
            )



"""
Core trace list implementation for pymseed

"""

from collections.abc import Sequence
from typing import Any, Optional, Callable

from .clib import clibmseed, ffi, cdata_to_string
from .definitions import DataEncoding, SubSecond, TimeFormat
from .exceptions import MiniSEEDError
from .msrecord import MS3Record
from .util import nstime2timestr, encoding_sizetype


class MS3RecordPtr:
    """Wrapper around CFFI MS3RecordPtr structure"""

    def __init__(self, cffi_ptr: Any) -> None:
        self._ptr = cffi_ptr

    def __repr__(self) -> str:
        return (
            f"Pointer to {self.msr.sourceid}, "
            f'{cdata_to_string(self._ptr.filename)}, '
            f"byte offset: {self._ptr.fileoffset}"
        )

    @property
    def msr(self) -> MS3Record:
        """Return a constructed MS3Record"""
        if not hasattr(self, "_msrecord"):
            self._msrecord = MS3Record(recordptr=self._ptr.msr)
        return self._msrecord

    @property
    def filename(self) -> Optional[str]:
        """Return filename as string"""
        result = cdata_to_string(self._ptr.filename)
        if result is None:
            return None
        return result

    @property
    def fileoffset(self) -> int:
        """Return file offset"""
        return self._ptr.fileoffset

    @property
    def endtime(self) -> int:
        """Return end time"""
        return self._ptr.endtime

    @property
    def dataoffset(self) -> int:
        """Return data offset"""
        return self._ptr.dataoffset


class MS3RecordList:
    """Wrapper around CFFI MS3RecordList structure"""

    def __init__(self, cffi_ptr: Any) -> None:
        self._list = cffi_ptr

    def __repr__(self) -> str:
        return f"Record list of {self._list.recordcnt} records"

    @property
    def recordcnt(self) -> int:
        """Return record count"""
        return self._list.recordcnt

    def records(self) -> Any:
        """Return the records via a generator iterator"""
        current_record = self._list.first
        while current_record != ffi.NULL:
            yield MS3RecordPtr(current_record)
            current_record = current_record.next


class MS3TraceSeg:
    """Wrapper around CFFI MS3TraceSeg structure"""

    def __init__(self, cffi_ptr: Any, parent_id_ptr: Any = None, parent_tracelist: Any = None) -> None:
        self._seg = cffi_ptr
        self._parent_id = parent_id_ptr  # Store reference to parent MS3TraceID
        self._parent_tracelist = parent_tracelist  # Store reference to parent MSTraceList

    def __repr__(self) -> str:
        return (
            f"start: {self.starttime_str()}, "
            f"end: {self.endtime_str()}, "
            f"samprate: {self.samprate}, "
            f"samples: {self.samplecnt} "
        )

    @property
    def starttime(self) -> int:
        """Return start time as nanoseconds since Unix/POSIX epoch"""
        return self._seg.starttime

    @property
    def starttime_seconds(self) -> float:
        """Return start time as seconds since Unix/POSIX epoch"""
        return self._seg.starttime / clibmseed.NSTMODULUS

    def starttime_str(
        self, timeformat: TimeFormat = TimeFormat.ISOMONTHDAY_Z, subsecond: SubSecond = SubSecond.NANO_MICRO_NONE
    ) -> Optional[str]:
        """Return start time as formatted string"""
        result = nstime2timestr(self._seg.starttime, timeformat, subsecond)
        if result is None:
            return None
        return result

    @property
    def endtime(self) -> int:
        """Return end time as nanoseconds since Unix/POSIX epoch"""
        return self._seg.endtime

    @property
    def endtime_seconds(self) -> float:
        """Return end time as seconds since Unix/POSIX epoch"""
        return self._seg.endtime / clibmseed.NSTMODULUS

    def endtime_str(
        self, timeformat: TimeFormat = TimeFormat.ISOMONTHDAY_Z, subsecond: SubSecond = SubSecond.NANO_MICRO_NONE
    ) -> Optional[str]:
        """Return end time as formatted string"""
        result = nstime2timestr(self._seg.endtime, timeformat, subsecond)
        if result is None:
            return None
        return result

    @property
    def samprate(self) -> float:
        """Return sample rate"""
        return self._seg.samprate

    @property
    def samplecnt(self) -> int:
        """Return sample count"""
        return self._seg.samplecnt

    @property
    def recordlist(self) -> Optional[MS3RecordList]:
        """Return the record list structure"""
        if self._seg.recordlist:
            return MS3RecordList(self._seg.recordlist)
        else:
            return None

    @property
    def datasamples(self) -> memoryview:
        """Return data samples as a memoryview (no copy)

        A view of the data samples in an internal buffer owned by this MS3Record
        instance is returned.  If the data are needed beyond the lifetime of this
        instance, a copy must be made.

        The returned view can be used directly with slicing and indexing
        from `0` to `MS3TraceSeg.numsamples - 1`.

        The view can efficiently be copied to a _python list_ using:

            data_samples = MS3TraceSeg.datasamples[:]
        """
        if self._seg.numsamples <= 0:
            return memoryview(b'')  # Empty memoryview

        sampletype = self.sampletype

        if sampletype == "i":
            ptr = ffi.cast("int32_t *", self._seg.datasamples)
            buffer = ffi.buffer(ptr, self._seg.numsamples * ffi.sizeof("int32_t"))
            return memoryview(buffer).cast('i')
        elif sampletype == "f":
            ptr = ffi.cast("float *", self._seg.datasamples)
            buffer = ffi.buffer(ptr, self._seg.numsamples * ffi.sizeof("float"))
            return memoryview(buffer).cast('f')
        elif sampletype == "d":
            ptr = ffi.cast("double *", self._seg.datasamples)
            buffer = ffi.buffer(ptr, self._seg.numsamples * ffi.sizeof("double"))
            return memoryview(buffer).cast('d')
        elif sampletype == "t":
            ptr = ffi.cast("char *", self._seg.datasamples)
            buffer = ffi.buffer(ptr, self._seg.numsamples)
            return memoryview(buffer).cast('B')
        else:
            raise ValueError(f"Unknown sample type: {sampletype}")

    @property
    def sampletype(self) -> Optional[str]:
        """Return sample type code if available, otherwise None"""
        if self._seg.sampletype:
            return str(self._seg.sampletype.decode("ascii"))
        else:
            return None

    @property
    def numsamples(self) -> int:
        """Return number of samples"""
        return self._seg.numsamples

    @property
    def datasize(self) -> int:
        """Return data size in bytes"""
        return self._seg.datasize

    @property
    def sample_size_type(self) -> tuple[int, str]:
        """Return data sample size and type code from first record in list

        NOTE: This is a guesstimate based on the first record in the record list.
        It is not guaranteed to be correct for any other records in the list.
        """
        if self._seg.recordlist is None:
            raise ValueError(
                "No record list available to determine sample size and type"
            )

        # Get the first record
        first_record_ptr = self._seg.recordlist.first

        if first_record_ptr is None:
            raise ValueError("No records in record list")

        return encoding_sizetype(first_record_ptr.msr.encoding)

    @property
    def np_datasamples(self) -> Any:
        """Return data samples as a numpy array (no copy)

        A view of the data samples in an internal buffer owned by this MS3TraceSeg
        instance is returned. If the data are needed beyond the lifetime of this
        instance, a copy must be made.
        """
        try:
            import numpy as np
        except ImportError:
            raise ImportError(
                "numpy is not installed. Install numpy or this package with [numpy] optional dependency"
            ) from None

        if self._seg.numsamples <= 0:
            return np.array([])  # Empty array

        sampletype = self.sampletype

        # Translate libmseed sample type to numpy type
        nptype = {
            "i": np.int32,
            "f": np.float32,
            "d": np.float64,
            "t": "S1",  # 1-byte strings for text data
        }

        if sampletype not in nptype:
            raise ValueError(f"Unknown sample type: {sampletype}")

        # Create numpy array view from CFFI buffer
        return np.frombuffer(self.datasamples, dtype=nptype[sampletype])

    def create_numpy_array_from_recordlist(self) -> Any:
        """Return data samples as a numpy array unpacked from the record list

        The numpy array returned is an independent copy of the data samples.
        """
        try:
            import numpy as np
        except ImportError:
            raise ImportError(
                "numpy is not installed. Install numpy or this package with [numpy] optional dependency"
            ) from None

        if self.recordlist is None:
            raise ValueError(
                "Record list required, use record_list=True when populating MS3TraceList"
            )

        if self.samplecnt <= 0:
            return np.array([])  # Empty array

        (_, sample_type) = self.sample_size_type

        # Translate libmseed sample type to numpy type
        nptype = {
            "i": np.int32,
            "f": np.float32,
            "d": np.float64,
            "t": "S1",  # 1-byte strings for text data
        }

        if sample_type not in nptype:
            raise ValueError(f"Unknown sample type: {sample_type}")

        # Create numpy array of the correct type and size
        array = np.empty(self.samplecnt, dtype=nptype[sample_type])

        # Unpack data samples into the array
        self.unpack_recordlist(buffer=array)

        return array

    def unpack_recordlist(self, buffer: Any = None, verbose: int = 0) -> int:
        """Unpack data samples from record list into a buffer-like object

        If a destination `buffer` is provided it must be a buffer-like object,
        and large enough to hold the data samples.

        If a destination `buffer` is not provided (None), unpacked data will be
        owned by this object instance and will be freed when the instance is
        destroyed. If you wish to keep the data, you must make a copy.

        Returns the number of samples unpacked.
        """
        if self.recordlist is None:
            raise ValueError("No record list available to unpack")

        if self.datasamples and buffer is not None:
            raise ValueError("Data samples already unpacked")

        # Handle buffer types that may not be compatible with memoryview
        buffer_ptr = ffi.NULL
        buffer_size = 0
        if buffer is not None:
            try:
                buffer_ptr = ffi.from_buffer(buffer)
                buffer_size = buffer.nbytes
            except (TypeError, AttributeError):
                # Try to get size through len() if nbytes is not available
                try:
                    buffer_ptr = ffi.from_buffer(buffer)
                    buffer_size = len(buffer) * buffer.itemsize if hasattr(buffer, 'itemsize') else len(buffer)
                except (TypeError, AttributeError):
                    raise ValueError("Buffer must support the buffer protocol")

        status = clibmseed.mstl3_unpack_recordlist(
            self._parent_id,
            self._seg,
            buffer_ptr,
            buffer_size,
            verbose,
        )

        if status < 0:
            raise MiniSEEDError(status, "Error unpacking record list")
        else:
            return status


class MS3TraceID:
    """Wrapper around CFFI MS3TraceID structure"""

    def __init__(self, cffi_ptr: Any, parent_tracelist: Any = None) -> None:
        self._id = cffi_ptr
        self._parent_tracelist = parent_tracelist

    def __repr__(self) -> str:
        return (
            f"Source ID: {self.sourceid}, "
            f"earliest: {self.earliest_str()}, "
            f"latest: {self.latest_str()}, "
            f"segments: {self.numsegments}"
        )

    @property
    def sourceid(self) -> Optional[str]:
        """Return source ID as string"""
        result = cdata_to_string(self._id.sid)
        if result is None:
            return None
        return result

    @property
    def pubversion(self) -> int:
        """Return publication version"""
        return self._id.pubversion

    @property
    def earliest(self) -> int:
        """Return earliest time as nanoseconds since Unix/POSIX epoch"""
        return self._id.earliest

    @property
    def earliest_seconds(self) -> float:
        """Return earliest time as seconds since Unix/POSIX epoch"""
        return self._id.earliest / clibmseed.NSTMODULUS

    def earliest_str(
        self, timeformat: TimeFormat = TimeFormat.ISOMONTHDAY_Z, subsecond: SubSecond = SubSecond.NANO_MICRO_NONE
    ) -> Optional[str]:
        """Return earliest time as formatted string"""
        result = nstime2timestr(self._id.earliest, timeformat, subsecond)
        if result is None:
            return None
        return result

    @property
    def latest(self) -> int:
        """Return latest time as nanoseconds since Unix/POSIX epoch"""
        return self._id.latest

    @property
    def latest_seconds(self) -> float:
        """Return latest time as seconds since Unix/POSIX epoch"""
        return self._id.latest / clibmseed.NSTMODULUS

    def latest_str(
        self, timeformat: TimeFormat = TimeFormat.ISOMONTHDAY_Z, subsecond: SubSecond = SubSecond.NANO_MICRO_NONE
    ) -> Optional[str]:
        """Return latest time as formatted string"""
        result = nstime2timestr(self._id.latest, timeformat, subsecond)
        if result is None:
            return None
        return result

    @property
    def numsegments(self) -> int:
        """Return number of segments"""
        return self._id.numsegments

    def segments(self) -> Any:
        """Return segments via a generator iterator"""
        current_segment = self._id.first
        while current_segment != ffi.NULL:
            yield MS3TraceSeg(current_segment, self._id, self._parent_tracelist)
            current_segment = current_segment.next


class _MS3TraceList:
    """Wrapper around CFFI MS3TraceList structure"""

    def __init__(self, cffi_ptr: Any) -> None:
        self._list = cffi_ptr

    @property
    def numtraceids(self) -> int:
        """Return number of trace IDs"""
        return self._list.numtraceids


class MS3TraceList:
    """A container for a list of traces read from miniSEED

    If `file_name` is specified miniSEED will be read from the file.

    If `unpack_data` is True, the data samples will be decoded.

    If `skip_not_data` is True, bytes from the input stream will be skipped
    until a record is found.

    If `validate_crc` is True, the CRC will be validated if contained in
    the record (legacy miniSEED v2 contains no CRCs).  The CRC provides an
    internal integrity check of the record contents.

    The overall structure of the trace list list of trace IDs, each of which
    contains a list of trace segments illustrated as follows:
    - MSTraceList
      - TraceID
        - Trace Segment
        - Trace Segment
        - Trace Segment
        - ...
      - TraceID
        - Trace Segment
        - Trace Segment
        - ...
      - ...

    MSTraceList.traces() returns a generator iterator for the list of TraceIDs,
    and TraceID.segments() returns a generator iterator for the list of trace
    segments.

    Example usage iterating over the trace list:
    ```
    from pymseed import MS3TraceList

    mstl = MS3TraceList('input_file.mseed')
    for traceid in mstl.traceids():
        print(f'{traceid.sourceid}, {traceid.pubversion}')
        for segment in traceid.segments():
            print(f'  {segment.starttime_str()} - {segment.endtime_str()}, ',
                  f'{segment.samprate} sps, {segment.samplecnt} samples')
    ```
    """

    def __init__(
        self,
        file_name=None,
        unpack_data=False,
        record_list=False,
        skip_not_data=False,
        validate_crc=True,
        split_version=False,
        verbose=0,
    ) -> None:
        # Initialize trace list - mstl3_init() returns an initialized pointer
        self._mstl = clibmseed.mstl3_init(ffi.NULL)

        if self._mstl == ffi.NULL:
            raise MiniSEEDError(clibmseed.MS_GENERROR, "Error initializing trace list")

        # Store filenames for record list functionality in C-like buffers
        self._c_file_names = []

        # Read specified file
        if file_name is not None:
            self.read_file(
                file_name,
                unpack_data,
                record_list,
                skip_not_data,
                validate_crc,
                split_version,
                verbose,
            )

    def __del__(self):
        """Destructor to ensure proper cleanup"""
        if self._mstl:
            mstl_ptr = ffi.new("MS3TraceList **")
            mstl_ptr[0] = self._mstl
            clibmseed.mstl3_free(mstl_ptr, 1)

    def __repr__(self) -> str:
        return f"MSTraceList with {self.numtraceids} trace IDs"

    @property
    def numtraceids(self) -> int:
        """Return number of trace IDs in the list"""
        if self._mstl == ffi.NULL:
            return 0
        return int(self._mstl.numtraceids)

    def get_traceid(self, sourceid: str, version: int = 0) -> Optional[MS3TraceID]:
        """Get a specific trace ID from the list"""
        c_sourceid = ffi.new("char[]", sourceid.encode("utf-8"))

        traceid_ptr = clibmseed.mstl3_findID(self._mstl, c_sourceid, version, ffi.NULL)

        if traceid_ptr == ffi.NULL:
            return None

        return MS3TraceID(traceid_ptr, self)

    def read_files(self, file_names: list[str], **kwargs: Any) -> None:
        """Read data from multiple files"""
        # Create C string array for file names
        _c_file_names: list[Any] = []
        for file_name in file_names:
            _c_file_names.append(ffi.new("char[]", file_name.encode("utf-8")))

        # Process each file
        for c_file_name in _c_file_names:
            self.read_file(ffi.string(c_file_name).decode("utf-8"), **kwargs)

    def traceids(self) -> Any:
        """Generator that yields MS3TraceID objects"""
        current_traceid = self._mstl.traces.next[0]
        while current_traceid != ffi.NULL:
            yield MS3TraceID(current_traceid, self)
            current_traceid = current_traceid.next[0]

    def sourceids(self) -> Any:
        """Return source IDs via a generator iterator"""
        for traceid in self.traceids():
            yield traceid.sourceid

    def print(
        self, details: int = 0, gaps: bool = False, versions: bool = False, timeformat: TimeFormat = TimeFormat.ISOMONTHDAY_Z
    ) -> None:
        """Print trace list details"""
        clibmseed.mstl3_printtracelist(self._mstl, timeformat, details, gaps, versions)

    def read_file(
        self,
        file_name: str,
        unpack_data: bool = False,
        record_list: bool = False,
        skip_not_data: bool = False,
        validate_crc: bool = True,
        split_version: bool = False,
        verbose: int = 0,
    ) -> None:
        """Read miniSEED data from file into trace list"""

        # Store files names for reference and use in record lists
        self._c_file_names.append(ffi.new("char[]", file_name.encode("utf-8")))
        c_file_name = self._c_file_names[-1]

        flags = 0
        if unpack_data:
            flags |= clibmseed.MSF_UNPACKDATA
        if record_list:
            flags |= clibmseed.MSF_RECORDLIST
        if skip_not_data:
            flags |= clibmseed.MSF_SKIPNOTDATA
        if validate_crc:
            flags |= clibmseed.MSF_VALIDATECRC

        # Create a reference to the current trace list pointer
        mstl_ptr = ffi.new("MS3TraceList **")
        mstl_ptr[0] = self._mstl

        status = clibmseed.ms3_readtracelist_selection(
            mstl_ptr,
            c_file_name,
            ffi.NULL,  # tolerance
            ffi.NULL,  # selections
            int(split_version),
            flags,
            verbose,
        )

        if status != clibmseed.MS_NOERROR:
            raise MiniSEEDError(status, f"Error reading file: {file_name}")

    def add_data(
        self,
        sourceid: str,
        data_samples: Sequence[Any],
        sample_type: str,
        sample_rate: float,
        start_time_str: Optional[str] = None,
        start_time: Optional[int] = None,
        start_time_seconds: Optional[float] = None,
        publication_version: int = 0,
    ) -> None:
        """Add data samples to the trace list"""

        # Create an MS3Record to hold the data
        msr = MS3Record()
        msr.sourceid = sourceid
        msr.samprate = sample_rate
        msr.pubversion = publication_version

        # Set start time
        if start_time_str is not None:
            msr.set_starttime_str(start_time_str)
        elif start_time is not None:
            msr.starttime = start_time
        elif start_time_seconds is not None:
            msr.starttime_seconds = start_time_seconds
        else:
            raise ValueError("Must specify one of start_time_str, start_time, or start_time_seconds")

        # Set sample count and type
        msr._msr.samplecnt = len(data_samples)
        msr._msr.numsamples = len(data_samples)

        if sample_type == "i":
            msr._msr.sampletype = b"i"
            msr.encoding = DataEncoding.INT32
        elif sample_type == "f":
            msr._msr.sampletype = b"f"
            msr.encoding = DataEncoding.FLOAT32
        elif sample_type == "d":
            msr._msr.sampletype = b"d"
            msr.encoding = DataEncoding.FLOAT64
        elif sample_type == "t":
            msr._msr.sampletype = b"t"
            msr.encoding = DataEncoding.TEXT
        else:
            raise ValueError(f"Unknown sample type: {sample_type}")

        # Allocate and copy data samples based on type
        if sample_type == "i":
            try:
                mv = memoryview(data_samples)
                if mv.format == 'i' and mv.itemsize == 4:
                    # Compatible format - safe to zero-copy
                    sample_array = ffi.cast("int32_t *", ffi.from_buffer(data_samples))
                else:
                    raise ValueError("Incompatible buffer format")
            except (TypeError, ValueError):
                # Not compatible or not a buffer - need conversion
                sample_array = ffi.new("int32_t[]", [int(sample) for sample in data_samples])
            msr._msr.datasamples = ffi.cast("void *", sample_array)
            msr._msr.datasize = len(data_samples) * 4
        elif sample_type == "f":
            try:
                mv = memoryview(data_samples)
                if mv.format == 'f' and mv.itemsize == 4:
                    # Compatible format - safe to zero-copy
                    sample_array = ffi.cast("float *", ffi.from_buffer(data_samples))
                else:
                    raise ValueError("Incompatible buffer format")
            except (TypeError, ValueError):
                # Not compatible or not a buffer - need conversion
                sample_array = ffi.new("float[]", [float(sample) for sample in data_samples])
            msr._msr.datasamples = ffi.cast("void *", sample_array)
            msr._msr.datasize = len(data_samples) * 4
        elif sample_type == "d":
            try:
                mv = memoryview(data_samples)
                if mv.format == 'd' and mv.itemsize == 8:
                    # Compatible format - safe to zero-copy
                    sample_array = ffi.cast("double *", ffi.from_buffer(data_samples))
                else:
                    raise ValueError("Incompatible buffer format")
            except (TypeError, ValueError):
                # Not compatible or not a buffer - need conversion
                sample_array = ffi.new("double[]", [float(sample) for sample in data_samples])
            msr._msr.datasamples = ffi.cast("void *", sample_array)
            msr._msr.datasize = len(data_samples) * 8
        elif sample_type == "t":
            try:
                mv = memoryview(data_samples)
                if mv.format in ('c', 'b', 'B') and mv.itemsize == 1:
                    # Compatible format - safe to zero-copy
                    sample_array = ffi.cast("char *", ffi.from_buffer(data_samples))
                else:
                    raise ValueError("Incompatible buffer format")
            except (TypeError, ValueError):
                # Not compatible or not a buffer - need conversion
                text_data = []
                for sample in data_samples:
                    if isinstance(sample, str):
                        text_data.append(sample.encode('utf-8')[0])
                    else:
                        text_data.append(int(sample) if isinstance(sample, (int, float)) else str(sample).encode('utf-8')[0])
                sample_array = ffi.new("char[]", text_data)
            msr._msr.datasamples = ffi.cast("void *", sample_array)
            msr._msr.datasize = len(data_samples)
        else:
            raise ValueError(f"Unknown sample type: {sample_type}")

        # Request storing time of update in the trace list segment (at seg.prvtptr)
        flags = clibmseed.MSF_PPUPDATETIME

        # Add the MS3Record to the trace list, setting auto-heal flag to 1 (true)
        segptr = clibmseed.mstl3_addmsr_recordptr(
            self._mstl, msr._msr, ffi.NULL, 0, 1, flags, ffi.NULL
        )

        # Disconnect data sample memory from the MS3Record
        msr._msr.datasamples = ffi.NULL

        if segptr == ffi.NULL:
            raise MiniSEEDError(clibmseed.MS_GENERROR, "Error adding data samples")

    def _record_handler_wrapper(self, record: Any, record_length: int, handlerdata: Any) -> None:
        """Callback function for mstl3_pack()"""
        # Convert CFFI buffer to bytes for the handler
        record_bytes = ffi.buffer(record, record_length)[:]
        self._record_handler(record_bytes, self._record_handler_data)

    def pack(
        self,
        handler: Callable[[bytes, Any], None],
        handlerdata: Any = None,
        flush_data: bool = True,
        record_length: int = 4096,
        encoding: DataEncoding = DataEncoding.STEIM1,
        format_version: Optional[int] = None,
        extra_headers: Optional[str] = None,
        verbose: int = 0,
    ) -> tuple[int, int]:
        """Pack data into miniSEED record(s) and call handler()

        Returns a tuple of (packed_samples, packed_records)
        """

        # Set handler function as CFFI callback function
        self._record_handler = handler
        self._record_handler_data = handlerdata

        # Create callback function type and instance
        RECORD_HANDLER = ffi.callback(
            "void(char *, int, void *)", self._record_handler_wrapper
        )

        pack_flags = 0
        if flush_data:
            pack_flags |= clibmseed.MSF_FLUSHDATA

        if format_version is not None:
            if format_version not in [2, 3]:
                raise ValueError(f"Invalid miniSEED format version: {format_version}")
            if format_version == 2:
                pack_flags |= clibmseed.MSF_PACKVER2

        packed_samples = ffi.new("int64_t *")

        c_extra = ffi.new("char[]", extra_headers.encode("utf-8")) if extra_headers else ffi.NULL

        packed_records = clibmseed.mstl3_pack(
            self._mstl,
            RECORD_HANDLER,
            ffi.NULL,
            record_length,
            encoding,
            packed_samples,
            pack_flags,
            verbose,
            c_extra,
        )

        if packed_records < 0:
            raise MiniSEEDError(packed_records, "Error packing miniSEED record(s)")

        return (packed_samples[0], packed_records)



#!/usr/bin/env python3
#
# Read miniSEED file(s) from a stream (stdin), select those that
# fall within the selected earliest and latest times, and write out
# to a stream (stdout).  Records that contain the selected times are
# trimmed to the selected times.
#
# Example usage:
#  cat input.mseed | stream_timewindow.py -e 2023-01-01T00:00:00 -l 2023-01-02T00:00:00 > output.mseed
#
# This file is part of the Python miniseed module.
# Copyright (c) 2025, EarthScope Data Services

import sys
import argparse
from pymseed import MS3RecordReader, MS3RecordBufferReader, timestr2nstime

# A global record buffer for repacked records
record_buffer = b""


def process_stream(args):
    records_written = 0
    bytes_written = 0

    print("Reading miniSEED from stdin, writing to stdout", file=sys.stderr)

    # Read miniSEED from stdin
    with MS3RecordReader(sys.stdin.fileno()) as msreader:
        for msr in msreader:

            # Skip records before earliest selection time
            if args.earliest and msr.endtime < args.earliest:
                continue

            # Skip records after latest selection time
            if args.latest and msr.starttime > args.latest:
                continue

            repacked_record = None

            # Trim record to selection time range if it contains selection times
            if (args.earliest and msr.starttime < args.earliest <= msr.endtime) or (
                args.latest and msr.starttime <= args.latest < msr.endtime
            ):
                repacked_record = trim_record(msr, args.earliest, args.latest)

            # Write either repacked or orignal record to stdout
            sys.stdout.buffer.write(repacked_record if repacked_record else msr.record)

            records_written += 1
            bytes_written += msr.reclen

    print(f"Wrote {records_written} records, {bytes_written} bytes", file=sys.stderr)


def record_handler(record, handler_data):
    """A callback function for MS3Record.set_record_handler()"""
    global record_buffer
    record_buffer = bytes(record)


def trim_record(msr, earliest, latest):
    """Trim a miniSEED record to the specified start and end times"""

    # Cannot trim time coverage of a record with no coverage
    if msr.samplecnt == 0 and msr.samprate == 0.0:
        return

    # Re-parse the record and decode the data samples
    buffer = bytearray(msr.record)  # Mutable/writable buffer required
    with MS3RecordBufferReader(buffer, unpack_data=True) as msreader:
        msr = msreader.read()

        starttime = msr.starttime
        endtime = msr.endtime
        sample_period_ns = int(1e9 / msr.samprate)
        datasamples = msr.datasamples[:]

        # Trim early samples to the earliest time
        if earliest and starttime < earliest <= endtime:
            count = 0
            while starttime < earliest:
                starttime += sample_period_ns
                count += 1

            msr.starttime = starttime
            datasamples = datasamples[count:]

        # Trim late samples to the latest time
        if latest and starttime <= latest < endtime:
            count = 0
            while endtime > latest:
                endtime -= sample_period_ns
                count += 1

            datasamples = datasamples[:-count]

        # Pack miniSEED record with handler writing to global record_buffer
        msr.pack(record_handler, datasamples=datasamples, sampletype=msr.sampletype)

        return record_buffer


def parse_timestr(timestr):
    """Helper for argparse to convert a time string to a nanosecond time value"""
    try:
        return timestr2nstime(timestr)
    except ValueError:
        raise argparse.ArgumentTypeError(f"Invalid time string: {timestr}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Streaming modifications of miniSEED")
    parser.add_argument(
        "--earliest",
        "-e",
        type=parse_timestr,
        help="Specify the earliest time to output",
    )
    parser.add_argument(
        "--latest", "-l", type=parse_timestr, help="Specify the latest time to output"
    )

    args = parser.parse_args()

    if args.earliest and args.latest and args.earliest > args.latest:
        parser.error("Earliest time is after latest time")

    process_stream(args)



#!/usr/bin/env python3
#
# Read miniSEED file(s) from a stream, accumulate stats, and write to
# a stream.  For this illustration input is stdin, output is stdout, and
# stats are printed to stderr on completion.
#
# Example usage:
#  cat input.mseed | stream_stats.py > output.mseed
#
# This file is part of the Python pymseed module.
# Copyright (c) 2025, EarthScope Data Services

import sys
import pprint
from collections import defaultdict
from pymseed import MS3RecordReader, nstime2timestr

# Container for trace stats
trace_stats = {}

print("Reading miniSEED from stdin, writing to stdout", file=sys.stderr)

# Read miniSEED from stdin
with MS3RecordReader(sys.stdin.fileno()) as msreader:
    for msr in msreader:

        stats = trace_stats.setdefault(msr.sourceid, defaultdict(int))

        stats["record_count"] += 1
        stats["sample_count"] += msr.samplecnt
        stats["bytes"] += msr.reclen

        # Track publication versions
        if msr.pubversion not in stats.setdefault("pubversions", []):
            stats["pubversions"].append(msr.pubversion)

        # Track format versions
        if msr.formatversion not in stats.setdefault("formatversions", []):
            stats["formatversions"].append(msr.formatversion)

        # Track earliest sample time
        if "earliest" not in stats or msr.starttime > stats["earliest"]:
            stats["earliest"] = msr.starttime

        # Track latest sample time
        if "latest" not in stats or msr.endtime > stats["latest"]:
            stats["latest"] = msr.endtime

        # Write raw miniSEED record to stdout
        sys.stdout.buffer.write(msr.record)

# Traverse trace stats and add date-time string values for earliest and latest
for stats in trace_stats.values():
    stats["earliest_str"] = nstime2timestr(stats["earliest"])
    stats["latest_str"] = nstime2timestr(stats["latest"])

# Pretty print stats to stderr (to avoid mixing with stdout)
pp = pprint.PrettyPrinter(stream=sys.stderr, indent=4, sort_dicts=False)
pp.pprint(trace_stats)
